import logging
import numpy as np
from imlib.general.system import get_num_processes


from cellfinder_core.classify.tools import get_model
from cellfinder_core.classify.cube_generator import CubeGeneratorFromFile
from cellfinder_core.train.train_yml import models


def main(
    points,
    signal_array,
    background_array,
    n_free_cpus,
    voxel_sizes,
    network_voxel_sizes,
    batch_size,
    cube_height,
    cube_width,
    cube_depth,
    trained_model,
    model_weights,
    network_depth,
    max_workers=3,
):

    # Too many workers doesn't increase speed, and uses huge amounts of RAM
    workers = get_num_processes(
        min_free_cpu_cores=n_free_cpus, n_max_processes=max_workers
    )

    logging.debug("Initialising cube generator")
    inference_generator = CubeGeneratorFromFile(
        points,
        signal_array,
        background_array,
        voxel_sizes,
        network_voxel_sizes,
        batch_size=batch_size,
        cube_width=cube_width,
        cube_height=cube_height,
        cube_depth=cube_depth,
    )

    model = get_model(
        existing_model=trained_model,
        model_weights=model_weights,
        network_depth=models[network_depth],
        inference=True,
    )

    logging.info("Running inference")
    predictions = model.predict(
        inference_generator,
        use_multiprocessing=True,
        workers=workers,
        verbose=True,
    )
    predictions = predictions.round()
    predictions = predictions.astype("uint16")

    predictions = np.argmax(predictions, axis=1)
    points_list = []

    # only go through the "extractable" points
    for idx, cell in enumerate(inference_generator.ordered_points):
        cell.type = predictions[idx] + 1
        points_list.append(cell)

    return points_list



"""
prep
==================
Functions to prepare files and directories needed for other functions
"""


import logging

from pathlib import Path

from imlib.general.system import get_num_processes

from imlib.general.config import get_config_obj

import cellfinder_core.tools.tf as tf_tools
from cellfinder_core.download import models as model_download
from cellfinder_core.download.download import amend_cfg
from cellfinder_core.tools.source_files import source_custom_config_cellfinder


def prep_classification(
    trained_model, model_weights, install_path, model, n_free_cpus
):
    n_processes = get_num_processes(min_free_cpu_cores=n_free_cpus)
    prep_tensorflow(n_processes)
    model_weights = prep_models(
        trained_model, model_weights, install_path, model
    )

    return model_weights


def prep_training(
    n_free_cpus, trained_model, model_weights, install_path, model
):
    n_processes = get_num_processes(min_free_cpu_cores=n_free_cpus)
    prep_tensorflow(n_processes)
    model_weights = prep_models(
        trained_model, model_weights, install_path, model
    )
    return model_weights


def prep_tensorflow(max_threads):
    tf_tools.set_tf_threads(max_threads)
    tf_tools.allow_gpu_memory_growth()


def prep_models(trained_model, model_weights, install_path, model):
    # if no model or weights, set default weights
    if trained_model is None and model_weights is None:
        logging.debug("No model or weights supplied, so using the default")

        config_file = source_custom_config_cellfinder()

        if not Path(config_file).exists():
            logging.debug("Custom config does not exist, downloading models")
            model_path = model_download.main(model, install_path)
            amend_cfg(new_model_path=model_path)

        model_weights = get_model_weights(config_file)
        if model_weights != "" and Path(model_weights).exists():
            model_weights = model_weights
        else:
            logging.debug("Model weights do not exist, downloading")
            model_path = model_download.main(model, install_path)
            amend_cfg(new_model_path=model_path)
            model_weights = get_model_weights(config_file)
    return model_weights


def get_model_weights(config_file):
    logging.debug(f"Reading config file: {config_file}")
    config_obj = get_config_obj(config_file)
    model_conf = config_obj["model"]
    model_weights = model_conf["model_path"]
    return model_weights



import glob
import os

from typing import Tuple

import numpy as np
from dask import array as da
from dask import delayed
from tifffile import TiffFile, imread

from imlib.general.system import get_sorted_file_paths


def get_tiff_meta(
    path: str,
) -> Tuple[Tuple[int, int], np.dtype]:
    with TiffFile(path) as tfile:
        nz = len(tfile.pages)
        if not nz:
            raise ValueError(f"tiff file {path} has no pages!")
        first_page = tfile.pages[0]

    return tfile.pages[0].shape, first_page.dtype


lazy_imread = delayed(imread)  # lazy reader


def read_with_dask(path):
    """
    Based on https://github.com/tlambert03/napari-ndtiffs
    :param path:
    :return:
    """

    filenames = glob.glob(os.path.join(path, "*.tif"))

    shape, dtype = get_tiff_meta(filenames[0])
    lazy_arrays = [lazy_imread(fn) for fn in get_sorted_file_paths(filenames)]
    dask_arrays = [
        da.from_delayed(delayed_reader, shape=shape, dtype=dtype)
        for delayed_reader in lazy_arrays
    ]
    stack = da.stack(dask_arrays, axis=0)
    return stack



"""
main
===============

Trains a network based on a yaml file specifying cubes of cells/non cells.

N.B imports are within functions to prevent tensorflow being imported before
it's warnings are silenced
"""


import os
import logging

from datetime import datetime
from pathlib import Path
from fancylog import fancylog
from argparse import (
    ArgumentParser,
    ArgumentDefaultsHelpFormatter,
    ArgumentTypeError,
)
from sklearn.model_selection import train_test_split
from imlib.general.numerical import check_positive_float, check_positive_int
from imlib.general.system import ensure_directory_exists
from imlib.IO.cells import find_relevant_tiffs
from imlib.IO.yaml import read_yaml_section

import cellfinder_core as program_for_log

home = Path.home()
install_path = home / ".cellfinder"

tf_suppress_log_messages = [
    "sample_weight modes were coerced from",
    "multiprocessing can interact badly with TensorFlow",
]

models = {
    "18": "18-layer",
    "34": "34-layer",
    "50": "50-layer",
    "101": "101-layer",
    "152": "152-layer",
}


def valid_model_depth(depth):
    """
    Ensures a correct existing_model is chosen
    :param value: Input value
    :param models: Dict of allowed models
    :return: Input value, if it corresponds to a valid existing_model
    """

    if depth in models.keys():
        return depth
    else:
        raise ArgumentTypeError(
            f"Model depth: {depth} is not valid. Please "
            f"choose one of: {list(models.keys())}"
        )


def misc_parse(parser):
    misc_parser = parser.add_argument_group("Misc options")
    misc_parser.add_argument(
        "--n-free-cpus",
        dest="n_free_cpus",
        type=check_positive_int,
        default=2,
        help="The number of CPU cores on the machine to leave "
        "unused by the program to spare resources.",
    )
    misc_parser.add_argument(
        "--max-ram",
        dest="max_ram",
        type=check_positive_float,
        default=None,
        help="Maximum amount of RAM to use (in GB) - not currently fully "
        "implemented for all parts of cellfinder",
    )
    misc_parser.add_argument(
        "--save-csv",
        dest="save_csv",
        action="store_true",
        help="Save .csv files of cell locations (in addition to xml)."
        "Useful for importing into other software.",
    )
    misc_parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Debug mode. Will increase verbosity of logging and save all "
        "intermediate files for diagnosis of software issues.",
    )
    misc_parser.add_argument(
        "--sort-input-file",
        dest="sort_input_file",
        action="store_true",
        help="If set to true, the input text file will be sorted using "
        "natural sorting. This means that the file paths will be "
        "sorted as would be expected by a human and "
        "not purely alphabetically",
    )
    return parser


def training_parse():
    from cellfinder_core.download.cli import (
        model_parser,
        download_directory_parser,
    )

    training_parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter
    )
    training_parser.add_argument(
        "-y",
        "--yaml",
        dest="yaml_file",
        nargs="+",
        required=True,
        type=str,
        help="The path to the yaml run file.",
    )
    training_parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        required=True,
        type=str,
        help="Output directory for the final model.",
    )
    training_parser.add_argument(
        "--continue-training",
        dest="continue_training",
        action="store_true",
        help="Continue training from an existing trained model. If no model "
        "or model weights are specified, this will continue from the "
        "included model.",
    )
    training_parser.add_argument(
        "--trained-model",
        dest="trained_model",
        type=str,
        help="Path to the trained model",
    )
    training_parser.add_argument(
        "--model-weights",
        dest="model_weights",
        type=str,
        help="Path to existing model weights",
    )
    training_parser.add_argument(
        "--network-depth",
        dest="network_depth",
        type=valid_model_depth,
        default="50",
        help="Resnet depth (based on He et al. (2015)",
    )
    training_parser.add_argument(
        "--batch-size",
        dest="batch_size",
        type=check_positive_int,
        default=16,
        help="Training batch size",
    )
    training_parser.add_argument(
        "--epochs",
        dest="epochs",
        type=check_positive_int,
        default=100,
        help="Number of training epochs",
    )
    training_parser.add_argument(
        "--test-fraction",
        dest="test_fraction",
        type=float,
        default=0.1,
        help="Fraction of training data to use for validation",
    )
    training_parser.add_argument(
        "--learning-rate",
        dest="learning_rate",
        type=check_positive_float,
        default=0.0001,
        help="Learning rate for training the model",
    )
    training_parser.add_argument(
        "--no-augment",
        dest="no_augment",
        action="store_true",
        help="Don't apply data augmentation",
    )
    training_parser.add_argument(
        "--save-weights",
        dest="save_weights",
        action="store_true",
        help="Only store the model weights, and not the full model. Useful to "
        "save storage space.",
    )
    training_parser.add_argument(
        "--no-save-checkpoints",
        dest="no_save_checkpoints",
        action="store_true",
        help="Store the model at intermediate points during training",
    )
    training_parser.add_argument(
        "--tensorboard",
        action="store_true",
        help="Log to output_directory/tensorboard",
    )
    training_parser.add_argument(
        "--save-progress",
        dest="save_progress",
        action="store_true",
        help="Save training progress to a .csv file",
    )

    training_parser = misc_parse(training_parser)
    training_parser = model_parser(training_parser)
    training_parser = download_directory_parser(training_parser)
    args = training_parser.parse_args()

    return args


def parse_yaml(yaml_files, section="data"):
    data = []
    for yaml_file in yaml_files:
        data.extend(read_yaml_section(yaml_file, section))
    return data


def get_tiff_files(yaml_contents):
    from cellfinder_core.tools.tiff import TiffList, TiffDir

    tiff_lists = []
    for d in yaml_contents:
        if d["bg_channel"] < 0:
            channels = [d["signal_channel"]]
        else:
            channels = [d["signal_channel"], d["bg_channel"]]
        if "cell_def" in d and d["cell_def"]:
            ch1_tiffs = [
                os.path.join(d["cube_dir"], f)
                for f in os.listdir(d["cube_dir"])
                if f.endswith("Ch" + str(channels[0]) + ".tif")
            ]
            tiff_lists.append(
                TiffList(
                    find_relevant_tiffs(ch1_tiffs, d["cell_def"]),
                    channels,
                    d["type"],
                )
            )
        else:
            tiff_lists.append(TiffDir(d["cube_dir"], channels, d["type"]))

    tiff_files = [tiff_dir.make_tifffile_list() for tiff_dir in tiff_lists]
    return tiff_files


def cli():
    args = training_parse()
    ensure_directory_exists(args.output_dir)

    fancylog.start_logging(
        args.output_dir,
        program_for_log,
        variables=[args],
        log_header="CELLFINDER TRAINING LOG",
    )

    output_dir = Path(args.output_dir)
    run(
        output_dir,
        args.yaml_file,
        n_free_cpus=args.n_free_cpus,
        trained_model=args.trained_model,
        model_weights=args.model_weights,
        install_path=args.install_path,
        model=args.model,
        network_depth=args.network_depth,
        learning_rate=args.learning_rate,
        continue_training=args.continue_training,
        test_fraction=args.test_fraction,
        batch_size=args.batch_size,
        no_augment=args.no_augment,
        tensorboard=args.tensorboard,
        save_weights=args.save_weights,
        no_save_checkpoints=args.no_save_checkpoints,
        save_progress=args.save_progress,
        epochs=args.epochs,
    )


def run(
    output_dir,
    yaml_file,
    n_free_cpus=2,
    trained_model=None,
    model_weights=None,
    install_path=install_path,
    model="resnet50_tv",
    network_depth="50",
    learning_rate=0.0001,
    continue_training=False,
    test_fraction=0.1,
    batch_size=16,
    no_augment=False,
    tensorboard=False,
    save_weights=False,
    no_save_checkpoints=False,
    save_progress=False,
    epochs=100,
):

    from cellfinder_core.main import suppress_tf_logging

    suppress_tf_logging(tf_suppress_log_messages)

    from tensorflow.keras.callbacks import (
        TensorBoard,
        ModelCheckpoint,
        CSVLogger,
    )

    from cellfinder_core.tools.prep import prep_training
    from cellfinder_core.classify.tools import make_lists, get_model
    from cellfinder_core.classify.cube_generator import CubeGeneratorFromDisk

    start_time = datetime.now()

    ensure_directory_exists(output_dir)
    model_weights = prep_training(
        n_free_cpus, trained_model, model_weights, install_path, model
    )

    yaml_contents = parse_yaml(yaml_file)

    tiff_files = get_tiff_files(yaml_contents)
    logging.info(
        f"Found {sum(len(imlist) for imlist in tiff_files)} images "
        f"from {len(yaml_contents)} datasets "
        f"in {len(yaml_file)} yaml files"
    )

    model = get_model(
        existing_model=trained_model,
        model_weights=model_weights,
        network_depth=models[network_depth],
        learning_rate=learning_rate,
        continue_training=continue_training,
    )

    signal_train, background_train, labels_train = make_lists(tiff_files)

    if test_fraction > 0:
        logging.info("Splitting data into training and validation datasets")
        (
            signal_train,
            signal_test,
            background_train,
            background_test,
            labels_train,
            labels_test,
        ) = train_test_split(
            signal_train,
            background_train,
            labels_train,
            test_size=test_fraction,
        )

        logging.info(
            f"Using {len(signal_train)} images for training and "
            f"{len(signal_test)} images for validation"
        )
        validation_generator = CubeGeneratorFromDisk(
            signal_test,
            background_test,
            labels=labels_test,
            batch_size=batch_size,
            train=True,
        )

        # for saving checkpoints
        base_checkpoint_file_name = "-epoch.{epoch:02d}-loss-{val_loss:.3f}.h5"

    else:
        logging.info("No validation data selected.")
        validation_generator = None
        base_checkpoint_file_name = "-epoch.{epoch:02d}.h5"

    training_generator = CubeGeneratorFromDisk(
        signal_train,
        background_train,
        labels=labels_train,
        batch_size=batch_size,
        shuffle=True,
        train=True,
        augment=not no_augment,
    )
    callbacks = []

    if tensorboard:
        logdir = output_dir / "tensorboard"
        ensure_directory_exists(logdir)
        tensorboard = TensorBoard(
            log_dir=logdir,
            histogram_freq=0,
            write_graph=True,
            update_freq="epoch",
        )
        callbacks.append(tensorboard)

    if not no_save_checkpoints:
        if save_weights:
            filepath = str(output_dir / ("weight" + base_checkpoint_file_name))
        else:
            filepath = str(output_dir / ("model" + base_checkpoint_file_name))

        checkpoints = ModelCheckpoint(
            filepath,
            save_weights_only=save_weights,
        )
        callbacks.append(checkpoints)

    if save_progress:
        filepath = str(output_dir / "training.csv")
        csv_logger = CSVLogger(filepath)
        callbacks.append(csv_logger)

    logging.info("Beginning training.")
    model.fit(
        training_generator,
        validation_data=validation_generator,
        use_multiprocessing=False,
        epochs=epochs,
        callbacks=callbacks,
    )

    if save_weights:
        logging.info("Saving model weights")
        model.save_weights(str(output_dir / "model_weights.h5"))
    else:
        logging.info("Saving model")
        model.save(output_dir / "model.h5")

    logging.info(
        "Finished training, " "Total time taken: %s",
        datetime.now() - start_time,
    )


if __name__ == "__main__":
    cli()



# distutils: language = c++

# cython: language_level=3

import cython

cimport libc.math as cmath

from libcpp.vector cimport vector as CppVector
from libcpp.map cimport map as CppMap
from libcpp.pair cimport pair as CppPair

import numpy as np

from cellfinder_core.detect.filters.typedefs cimport ull, uint, Point

from imlib.cells.cells import Cell



DEF ULLONG_MAX = 18446744073709551615 # (2**64) -1
DEF N_NEIGHBOURS_4_CONNECTED = 3  # top left, below
DEF N_NEIGHBOURS_8_CONNECTED = 13  # all the 9 below + the 4 before on same plane



cdef get_non_zero_ull_min(ull[:] values):
    cdef ull min_val = ULLONG_MAX
    cdef ull s_id
    cdef uint i
    for i in range(len(values)):
        s_id = values[i]
        if s_id != 0:
            if s_id < min_val:
                min_val = s_id
    return min_val

cpdef get_non_zero_ull_min_wrapper(values):  # wrapper for testing purposes
    assert len(values) == 10
    cdef ull c_values[10]
    for i, v in enumerate(values):
        c_values[i] = <ull> v
    return get_non_zero_ull_min(c_values)

cdef get_structure_centre(CppVector[Point] structure):
    cdef double mean_z, mean_y, mean_x
    mean_x = 0; mean_y = 0; mean_z = 0
    cdef double s_len = len(structure)

    cdef Point p
    for p in structure:
        mean_x += p.x / s_len
        mean_y += p.y / s_len
        mean_z += p.z / s_len

    return Point(<ull> cmath.round(mean_x), <ull> cmath.round(mean_y), <ull> cmath.round(mean_z))


cpdef get_structure_centre_wrapper(structure):  # wrapper for testing purposes
    cdef CppVector[Point] s
    for p in structure:
        if type(p) == dict:
            s.push_back(Point(p['x'], p['y'], p['z']))
        else:
            s.push_back(Point(p.x, p.y, p.z))
    return get_structure_centre(s)



cdef class CellDetector:
    cdef:
        public ull SOMA_CENTRE_VALUE  # = range - 1 (e.g. 2**16 - 1)

        ull z
        int relative_z
        int connect_type

        ull[:,:] previous_layer
        tuple shape

        StructureManager structure_manager

        ull next_structure_id

    def __init__(self, uint width, uint height, uint start_z, connect_type=4):
        self.shape = width, height
        self.z = start_z

        assert connect_type in (4, 8), 'Connection type must be one of 4,8 got "{}"'.format(connect_type)
        self.connect_type = connect_type

        self.SOMA_CENTRE_VALUE = ULLONG_MAX

        self.relative_z = 0  # position to append in stack  # FIXME: replace by keeping start_z and self.z > self.start_Z
        self.next_structure_id = 1

        self.structure_manager = StructureManager()

    cpdef get_previous_layer(self):
        return np.array(self.previous_layer, dtype=np.uint64)

    cpdef process(self, layer):  # WARNING: inplace  # WARNING: ull may be overkill but ulong required
        assert [e for e in layer.shape[:2]] == [e for e in self.shape], \
            'CellDetector layer error, expected shape "{}", got "{}"'\
                .format(self.shape, [e for e in layer.shape[:2]])

        source_dtype = layer.dtype
        layer = layer.astype(np.uint64)

        cdef ull[:,:] c_layer

        if source_dtype == np.uint8:
            layer *= 72340172838076656  # TEST:
        elif source_dtype == np.uint16:
            layer *= 281479271743489
        elif source_dtype == np.uint32:
            layer *= 4294967297
        elif source_dtype == np.uint64:
            pass
        else:
            raise ValueError('Expected layer of any type from np.uint8, np.uint16, np.uint32, np.uint64,'
                             'got: {}'.format(source_dtype))
        c_layer = layer

        if self.connect_type == 4:
            self.previous_layer = self.connect_four(c_layer)
        else:
            self.previous_layer = self.connect_eight(c_layer)

        if self.relative_z == 0:
            self.relative_z += 1

        self.z += 1

    @cython.boundscheck(False)
    cdef connect_four(self, ull[:,:] layer):
        """
        For all the pixels in the current layer, finds all structures touching this pixel using the
        four connected (plus shape) rule and also looks at the pixel at the same location in the previous layer.
        If structures are found, they are added to the structure manager and the pixel labeled accordingly.

        :param layer:
        :return:
        """
        cdef ull struct_id
        cdef ull neighbour_ids[N_NEIGHBOURS_4_CONNECTED]
        cdef uint i
        for i in range(N_NEIGHBOURS_4_CONNECTED):  # reset
            neighbour_ids[i] = 0 # Labels of structures at left, top, below

        cdef uint y, x
        for y in range(layer.shape[1]):
            for x in range(layer.shape[0]):
                if layer[x, y] == self.SOMA_CENTRE_VALUE:
                    for i in range(N_NEIGHBOURS_4_CONNECTED):  # reset
                        neighbour_ids[i] = 0 # Labels of structures at left, top, below
                    # If in bounds look at neighbours
                    if x > 0:
                        neighbour_ids[0] = layer[x-1, y]
                    if y > 0:
                        neighbour_ids[1] = layer[x, y-1]
                    if self.relative_z > 0:
                        neighbour_ids[2] = self.previous_layer[x, y]

                    if self.is_new_structure(neighbour_ids):
                        neighbour_ids[0] = self.next_structure_id
                        self.next_structure_id += 1
                    struct_id = self.structure_manager.add(x, y, self.z, neighbour_ids)
                else:
                    struct_id = 0  # reset so that grayscale value does not count as structure in next iterations

                layer[x, y] = struct_id
        return layer

    cdef connect_eight(self, ull[:,:] layer):
        """
        For all the pixels in the current layer, finds all structures touching this pixel using the
        eight connected (connected by edges or corners) rule and also looks at the pixel at the same
        location in the previous layer.
        If structures are found, they are added to the structure manager and the pixel labeled accordingly.

        :param layer:
        :return:
        """
        cdef ull struct_id
        cdef ull neighbour_ids[N_NEIGHBOURS_8_CONNECTED]
        cdef uint i
        for i in range(N_NEIGHBOURS_8_CONNECTED):  # reset
            neighbour_ids[i] = 0  # Labels of neighbour structures touching before

        cdef uint y, x
        for y in range(layer.shape[1]):
            for x in range(layer.shape[0]):
                if layer[x, y] == self.SOMA_CENTRE_VALUE:
                    for i in range(N_NEIGHBOURS_8_CONNECTED):  # reset
                        neighbour_ids[i] = 0

                    # If in bounds look at neighbours
                    if x > 0 and y > 0:
                        neighbour_ids[0] = layer[x-1, y-1]
                    if x > 0:
                        neighbour_ids[1] = layer[x-1, y]
                    if y > 0:
                        neighbour_ids[2] = layer[x, y-1]
                        neighbour_ids[3] = layer[x+1, y-1]
                    if self.relative_z > 0:
                        if x > 0 and y > 0:
                            neighbour_ids[4] = self.previous_layer[x-1, y-1]
                        if x > 0:
                            neighbour_ids[5] = self.previous_layer[x-1, y]
                            if y < layer.shape[1] - 1:
                                neighbour_ids[6] = self.previous_layer[x-1, y+1]
                        if y > 0:
                            neighbour_ids[7] = self.previous_layer[x, y-1]
                            if x < layer.shape[0] - 1:
                                neighbour_ids[8] = self.previous_layer[x+1, y-1]
                        neighbour_ids[9] = self.previous_layer[x, y]
                        if y < layer.shape[1] - 1:
                            neighbour_ids[10] = self.previous_layer[x, y+1]
                        if x < layer.shape[0] - 1:
                            neighbour_ids[11] = self.previous_layer[x+1, y]
                            if y < layer.shape[1] - 1:
                                neighbour_ids[12] = self.previous_layer[x+1, y+1]

                    if self.is_new_structure(neighbour_ids):
                        neighbour_ids[0] = self.next_structure_id
                        self.next_structure_id += 1
                    struct_id = self.structure_manager.add(x, y, self.z, neighbour_ids)
                else:
                    struct_id = 0  # reset so that grayscale value does not count as structure in next iterations

                layer[x, y] = struct_id
        return layer

    @cython.boundscheck(False)
    cdef is_new_structure(self, ull[:] neighbour_ids):  # TEST:
        cdef uint i
        for i in range(len(neighbour_ids)):
            if neighbour_ids[i] != 0:
                return False
        return True

    cpdef get_cell_centres(self):
        cdef CppVector[Point] cell_centres
        cell_centres = self.structure_manager.structures_to_cells()
        return cell_centres

    cpdef get_coords_list(self):
        coords = self.structure_manager.get_coords_dict()  # TODO: cache (attribute)
        return coords

cdef class StructureManager:
    cdef:
        CppMap[ull, ull] obsolete_ids
        CppMap[ull, CppVector[Point]] coords_maps
        int default_cell_type

    def __init__(self):
        self.default_cell_type = Cell.UNKNOWN

    cpdef get_coords_dict(self):
        return self.coords_maps

    @cython.boundscheck(False)
    cdef add(self, uint x, uint y, uint z, ull[:] neighbour_ids):
        """
        For the current coordinates takes all the neighbours and find the minimum structure
        including obsolete structures mapping to any of the neighbours recursively.
        Once the correct structure id is found, append a point with the current coordinates to the coordinates map
        entry for the correct structure. Hence each entry of the map will be a vector of all the pertaining points.

        :param x: 
        :param y: 
        :param z: 
        :param neighbour_ids: 
        :return: 
        """
        cdef ull updated_id

        updated_id = self.sanitise_ids(neighbour_ids)
        self.merge_structures(updated_id, neighbour_ids)

        cdef Point p = Point(x, y, z)  # Necessary to split definition on some machines
        self.coords_maps[updated_id].push_back(p)  # Add point for that structure

        return updated_id

    @cython.boundscheck(False)
    cdef sanitise_ids(self, ull[:] neighbour_ids):
        """
        For all the neighbour ids, walk up the chain of obsolescence (self.obsolete_ids)
        to reassign the corresponding most obsolete structure to the current neighbour

        :param neighbour_ids:
        :return: updated_id
        """
        cdef ull updated_id, neighbour_id
        cdef uint i
        for i in range(len(neighbour_ids)):
            neighbour_id = neighbour_ids[i]
            while self.obsolete_ids.count(neighbour_id):  # walk up the chain of obsolescence
                neighbour_id = self.obsolete_ids[neighbour_id]
            neighbour_ids[i] = neighbour_id

        updated_id = get_non_zero_ull_min(neighbour_ids)  # FIXME: what happens if all neighbour_ids are 0 (raise)
        return updated_id

    @cython.boundscheck(False)
    cdef merge_structures(self, ull updated_id, ull[:] neighbour_ids):
        """
        For all the neighbours, reassign all the points of neighbour to updated_id
        Then deletes the now obsolete entry from the points map and add that entry to the obsolete_ids

        :param updated_id:
        :param neighbour_ids:
        """
        cdef ull neighbour_id
        cdef Point p
        cdef uint i
        for i in range(len(neighbour_ids)):
            neighbour_id = neighbour_ids[i]
            if neighbour_id > updated_id:  # minimise ID so if neighbour with higher ID, reassign its points to current
                for p in self.coords_maps[neighbour_id]:
                    self.coords_maps[updated_id].push_back(p)
                self.coords_maps.erase(neighbour_id)
                self.obsolete_ids[neighbour_id] = updated_id

    cdef structures_to_cells(self):
        cdef CppVector[Point] structure, cell_centres
        cdef Point p

        cdef CppPair[ull, CppVector[Point]] iterator_pair
        for iterator_pair in self.coords_maps:
            structure = iterator_pair.second
            p = get_structure_centre(structure)
            cell_centres.push_back(p)
        return cell_centres



import logging

import numpy as np
import tensorflow as tf

from cellfinder_core.classify.resnet import build_model


def get_model(
    existing_model=None,
    model_weights=None,
    network_depth=None,
    learning_rate=0.0001,
    inference=False,
    continue_training=False,
):
    """
    Returns the correct model based on the arguments passed
    :param existing_model: An existing, trained model. This is returned if it
    exists
    :param model_weights: This file is used to set the model weights if it
    exists
    :param network_depth: This defines the type of model to be created if
    necessary
    :param learning_rate: For creating a new model
    :param inference: If True, will ensure that a trained model exists. E.g.
    by using the default one
    :param continue_training: If True, will ensure that a trained model
    exists. E.g. by using the default one
    :return: A tf.keras model

    """
    if existing_model is not None:
        logging.debug(f"Loading model: {existing_model}")
        return tf.keras.models.load_model(existing_model)
    else:
        logging.debug(f"Creating a new instance of model: {network_depth}")
        model = build_model(
            network_depth=network_depth, learning_rate=learning_rate
        )
        if inference or continue_training:
            logging.debug(
                f"Setting model weights according to: {model_weights}"
            )
            model.load_weights(model_weights)
        return model


def make_lists(tiff_files, train=True):

    signal_list = []
    background_list = []
    if train:
        labels = []

    for group in tiff_files:
        for image in group:
            signal_list.append(image.img_files[0])
            background_list.append(image.img_files[1])
            if train:
                if image.label == "no_cell":
                    labels.append(0)
                elif image.label == "cell":
                    labels.append(1)

    if train:
        labels = np.array(labels)
        return signal_list, background_list, labels
    else:
        return signal_list, background_list



from tensorflow.keras import Model
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.layers import (
    Input,
    ZeroPadding3D,
    Conv3D,
    Activation,
    MaxPooling3D,
    GlobalAveragePooling3D,
    Dense,
    BatchNormalization,
    Add,
)

#####################################################################
# Define the types of ResNet

resnet_unit_blocks = {
    "18-layer": [2, 2, 2, 2],
    "34-layer": [3, 4, 6, 3],
    "50-layer": [3, 4, 6, 3],
    "101-layer": [3, 4, 23, 3],
    "152-layer": [3, 6, 36, 3],
}

network_residual_bottleneck = {
    "18-layer": False,
    "34-layer": False,
    "50-layer": True,
    "101-layer": True,
    "152-layer": True,
}
#####################################################################


def build_model(
    shape=(50, 50, 20, 2),
    network_depth="18-layer",
    optimizer=None,
    learning_rate=0.0005,  # higher rates don't always converge
    loss="categorical_crossentropy",
    metrics=["accuracy"],
    number_classes=2,
    axis=3,
    starting_features=64,
    classification_activation="softmax",
):
    """

    :param shape:
    :param network_depth:
    :param optimizer:
    :param learning_rate:
    :param loss:
    :param metrics:
    :param number_classes:
    :param int axis: Default: 3. Assumed channels are last
    :param starting_features: # increases in each set of residual units
    :param classification_activation:
    :return:
    """
    blocks, bottleneck = get_resnet_blocks_and_bottleneck(network_depth)

    inputs = Input(shape)
    x = non_residual_block(inputs, starting_features, axis=axis)

    features = starting_features
    for resnet_unit_id, iterations in enumerate(blocks):
        for block_id in range(iterations):
            x = residual_block(
                features,
                resnet_unit_id,
                block_id,
                bottleneck=bottleneck,
                axis=axis,
            )(x)

        features *= 2

    x = GlobalAveragePooling3D(name="final_average_pool")(x)
    x = Dense(
        number_classes,
        activation=classification_activation,
        name="fully_connected",
    )(x)

    # Instantiate existing_model.
    model = Model(inputs=inputs, outputs=x)

    if optimizer is None:
        optimizer = Adam(learning_rate=learning_rate)

    model.compile(optimizer, loss=loss, metrics=metrics)
    return model


def get_resnet_blocks_and_bottleneck(network_depth):
    """
    Parses dicts, and returns how many resnet blocks are in each unit, along
    with whether they are bottlneck blocks or not

    :param network_depth:
    :return:
    """
    blocks = resnet_unit_blocks[network_depth]
    bottleneck = network_residual_bottleneck[network_depth]
    return blocks, bottleneck


def non_residual_block(
    inputs,
    starting_features,
    conv_kernel=(7, 7, 3),
    strides=(2, 2, 2),
    padding=3,
    max_pool_size=(3, 3, 2),
    activation="relu",
    use_bias=False,
    bn_epsilon=1e-5,
    pooling_padding="same",
    axis=3,
):
    """
     Non-residual unit from He et al. (2015). Corresponds to "conv1" and the
    max pool.

    :param inputs:
    :param starting_features:
    :param conv_kernel:
    :param strides:
    :param padding:
    :param max_pool_size:
    :param activation:
    :param use_bias:
    :param bn_epsilon:
    :param pooling_padding:
    :param axis:
    :return:
    """

    x = ZeroPadding3D(padding=padding, name="conv1_padding")(inputs)
    x = Conv3D(
        starting_features,
        conv_kernel,
        strides=strides,
        use_bias=use_bias,
        name="conv1",
    )(x)
    x = BatchNormalization(axis=axis, epsilon=bn_epsilon, name="conv1_bn")(x)
    x = Activation(activation, name="conv1_activation")(x)
    x = MaxPooling3D(
        max_pool_size,
        strides=strides,
        padding=pooling_padding,
        name="max_pool",
    )(x)

    return x


def residual_block(
    output_features,
    resnet_unit_id,
    block_id,
    conv_kernel=(3, 3, 3),
    bottleneck_conv_kernel=(1, 1, 1),
    bottleneck=False,
    activation="relu",
    use_bias=False,
    kernel_initializer="he_normal",
    bn_epsilon=1e-5,
    axis=3,
):

    """
    Residual unit from He et al. (2015)


    :param int output_features: How many output features
    :param int resnet_unit_id: Which resnet unit
    (e.g. 0 is conv2_x, 1 is conv3_x)
    :param int block_id: Which block in the resnet unit (i.e. in the third
    unit of a 152-layer resnet, block_id=36)
    :param conv_kernel:
    :param bottleneck_conv_kernel:
    :param bool bottleneck: If True, use a bottleneck resnet unit.
    Default False.
    :param activation:
    :param use_bias:
    :param kernel_initializer:
    :param bn_epsilon:
    :param axis:
    :return:
    """

    stride = get_stride(resnet_unit_id, block_id)
    resnet_unit_label = resnet_unit_id + 2

    def f(x):
        if bottleneck:
            y = Conv3D(
                output_features,
                bottleneck_conv_kernel,
                strides=stride,
                use_bias=use_bias,
                name=f"resunit{resnet_unit_label}_block{block_id}_conv_a",
                kernel_initializer=kernel_initializer,
            )(x)
        else:
            y = ZeroPadding3D(
                padding=1,
                name=f"resunit{resnet_unit_label}_block{block_id}_pad_a",
            )(x)

            y = Conv3D(
                output_features,
                conv_kernel,
                strides=stride,
                use_bias=use_bias,
                name=f"resunit{resnet_unit_label}_block{block_id}_conv_a",
                kernel_initializer=kernel_initializer,
            )(y)

        y = BatchNormalization(
            axis=axis,
            epsilon=bn_epsilon,
            name=f"resunit{resnet_unit_label}_block{block_id}_bn_a",
        )(y)

        y = Activation(
            activation,
            name=f"resunit{resnet_unit_label}_block{block_id}_activation_a",
        )(y)

        y = ZeroPadding3D(
            padding=1, name=f"resunit{resnet_unit_label}_block{block_id}_pad_b"
        )(y)

        y = Conv3D(
            output_features,
            conv_kernel,
            use_bias=use_bias,
            name=f"resunit{resnet_unit_label}_block{block_id}_conv_b",
            kernel_initializer=kernel_initializer,
        )(y)

        y = BatchNormalization(
            axis=axis,
            epsilon=bn_epsilon,
            name=f"resunit{resnet_unit_label}_block{block_id}_bn_b",
        )(y)

        if bottleneck:
            y = Activation(
                activation,
                name=f"resunit{resnet_unit_label}_block"
                f"{block_id}_activation_b",
            )(y)

            y = Conv3D(
                output_features * 4,
                bottleneck_conv_kernel,
                use_bias=use_bias,
                name=f"resunit{resnet_unit_label}_block{block_id}_conv_c",
                kernel_initializer=kernel_initializer,
            )(y)

            y = BatchNormalization(
                axis=axis,
                epsilon=bn_epsilon,
                name=f"resunit{resnet_unit_label}_block{block_id}_bn_c",
            )(y)

            identity_shortcut = get_shortcut(
                x,
                resnet_unit_label,
                block_id,
                output_features * 4,
                stride,
                axis=axis,
            )
        else:
            identity_shortcut = get_shortcut(
                x,
                resnet_unit_label,
                block_id,
                output_features,
                stride,
                axis=axis,
            )

        y = Add(name=f"resunit{resnet_unit_label}_block{block_id}_add")(
            [y, identity_shortcut]
        )

        y = Activation(
            activation,
            name=f"resunit{resnet_unit_label}_block{block_id}_activation_c",
        )(y)

        return y

    return f


def get_shortcut(
    inputs,
    resnet_unit_label,
    block_id,
    features,
    stride,
    use_bias=False,
    kernel_initializer="he_normal",
    bn_epsilon=1e-5,
    axis=3,
):

    """
    Create shortcut. For none-bottleneck residual units, this is just the
    identity. Otherwise, the input is reshaped to match the output of the
    bottleneck unit

    :param inputs: Input to the residual unit
    :param int resnet_unit_label: Which resnet unit (i.e. 2 is conv2_x)
    :param int block_id: Which block in the resnet unit (i.e. in the third
    unit of a 152-layer resnet, block_id=36)
    :param int features: How many output features
    :param stride: Convolution stride
    :param use_bias:
    :param kernel_initializer:
    :param bn_epsilon:
    :param axis:
    :return: Shortcut tensor, to add to the output of the residual unit
    """

    if block_id == 0:
        shortcut = Conv3D(
            features,
            (1, 1, 1),
            strides=stride,
            use_bias=use_bias,
            name=f"resunit{resnet_unit_label}_block{block_id}_shortcut_conv",
            kernel_initializer=kernel_initializer,
        )(inputs)

        shortcut = BatchNormalization(
            axis=axis,
            epsilon=bn_epsilon,
            name=f"resunit{resnet_unit_label}_block{block_id}_shortcut_bn",
        )(shortcut)
        return shortcut
    else:
        return inputs


def get_stride(resnet_unit_id, block_id):
    """
    Determines the convolution stride.

    :param int resnet_unit_id: Which resnet unit
    (e.g. 0 is conv2_x, 1 is conv3_x)
    :param int block_id: Which block in the resnet unit (i.e. in the third
    unit of a 152-layer resnet, block_id=36)
    :return: Stride
    """
    if resnet_unit_id == 0 or block_id != 0:
        return 1
    else:
        return 2



import platform
from setuptools import setup, find_packages, Extension
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "numpy",
    "scikit-learn",
    "configparser",
    "pandas",
    "packaging",
    "scikit-image",
    "tifffile",
    "natsort",
    "tqdm",
    "multiprocessing-logging",
    "psutil",
    "configobj",
    "tensorflow>=2.4.0",
    "fancylog>=0.0.7",
    "imlib>=0.0.26",
    "slurmio>=0.0.4",
    "imio",
    "dask[array]",
]


if platform.system() == "Windows":
    # FIXME: There must be a better way of doing this.
    base_tile_filter_extension = Extension(
        name="cellfinder_core.detect.filters.plane.base_tile_filter",
        sources=["cellfinder_core/detect/filters/plane/base_tile_filter.pyx"],
        language="c++",
    )

    ball_filter_extension = Extension(
        name="cellfinder_core.detect.filters.volume.ball_filter",
        sources=["cellfinder_core/detect/filters/volume/ball_filter.pyx"],
    )

    structure_detection_extension = Extension(
        name="cellfinder_core.detect.filters.volume.structure_detection",
        sources=[
            "cellfinder_core/detect/filters/volume/structure_detection.pyx"
        ],
        language="c++",
    )
else:
    base_tile_filter_extension = Extension(
        name="cellfinder_core.detect.filters.plane.base_tile_filter",
        sources=["cellfinder_core/detect/filters/plane/base_tile_filter.pyx"],
        libraries=["m"],
        language="c++",
    )

    ball_filter_extension = Extension(
        name="cellfinder_core.detect.filters.volume.ball_filter",
        sources=["cellfinder_core/detect/filters/volume/ball_filter.pyx"],
        libraries=["m"],
    )

    structure_detection_extension = Extension(
        name="cellfinder_core.detect.filters.volume.structure_detection",
        sources=[
            "cellfinder_core/detect/filters/volume/structure_detection.pyx"
        ],
        libraries=["m"],
        language="c++",
    )


setup(
    name="cellfinder-core",
    version="0.2.3-rc0",
    description="Automated 3D cell detection in large microscopy images",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    extras_require={
        "dev": [
            "black",
            "pytest-cov",
            "pytest",
            "gitpython",
            "coverage>=5.0.3",
            "bump2version",
            "pre-commit",
            "flake8",
        ]
    },
    setup_requires=["cython"],
    python_requires=">=3.7",
    packages=find_packages(),
    include_package_data=True,
    ext_modules=[
        ball_filter_extension,
        structure_detection_extension,
        base_tile_filter_extension,
    ],
    entry_points={
        "console_scripts": [
            "cellfinder_download = cellfinder_core.download.cli:main",
            "cellfinder_train = cellfinder_core.train.train_yml:cli",
            # "cellfinder_curate_new = cellfinder_core.train.curation:main",
            # "cellfinder_curate = cellfinder_core.train.curation_old:main",
        ]
    },
    url="https://cellfinder.info",
    project_urls={
        "Source Code": "https://github.com/brainglobe/cellfinder-core",
        "Bug Tracker": "https://github.com/brainglobe/cellfinder-core/issues",
        "Documentation": "https://docs.brainglobe.info/cellfinder",
    },
    author="Adam Tyson, Christian Niedworok, Charly Rousseau",
    author_email="code@adamltyson.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    zip_safe=False,
)



from cpython.object cimport PyObject

cdef extern from "Python.h":
    ctypedef struct PyTypeObject:
        pass

cdef extern from "datetime.h":

    ctypedef extern class datetime.date[object PyDateTime_Date]:
        pass

    ctypedef extern class datetime.time[object PyDateTime_Time]:
        pass

    ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
        pass

    ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
        pass

    ctypedef extern class datetime.tzinfo[object PyDateTime_TZInfo]:
        pass

    ctypedef struct PyDateTime_Date:
        pass

    ctypedef struct PyDateTime_Time:
        char hastzinfo
        PyObject *tzinfo

    ctypedef struct PyDateTime_DateTime:
        char hastzinfo
        PyObject *tzinfo

    ctypedef struct PyDateTime_Delta:
        int days
        int seconds
        int microseconds

    # Define structure for C API.
    ctypedef struct PyDateTime_CAPI:
        # type objects
        PyTypeObject *DateType
        PyTypeObject *DateTimeType
        PyTypeObject *TimeType
        PyTypeObject *DeltaType
        PyTypeObject *TZInfoType

        # constructors
        object (*Date_FromDate)(int, int, int, PyTypeObject*)
        object (*DateTime_FromDateAndTime)(int, int, int, int, int, int, int, object, PyTypeObject*)
        object (*Time_FromTime)(int, int, int, int, object, PyTypeObject*)
        object (*Delta_FromDelta)(int, int, int, int, PyTypeObject*)

        # constructors for the DB API
        object (*DateTime_FromTimestamp)(object, object, object)
        object (*Date_FromTimestamp)(object, object)

    # Check type of the object.
    bint PyDate_Check(object op)
    bint PyDate_CheckExact(object op)

    bint PyDateTime_Check(object op)
    bint PyDateTime_CheckExact(object op)

    bint PyTime_Check(object op)
    bint PyTime_CheckExact(object op)

    bint PyDelta_Check(object op)
    bint PyDelta_CheckExact(object op)

    bint PyTZInfo_Check(object op)
    bint PyTZInfo_CheckExact(object op)

    # Getters for date and datetime (C macros).
    int PyDateTime_GET_YEAR(object o)
    int PyDateTime_GET_MONTH(object o)
    int PyDateTime_GET_DAY(object o)

    # Getters for datetime (C macros).
    int PyDateTime_DATE_GET_HOUR(object o)
    int PyDateTime_DATE_GET_MINUTE(object o)
    int PyDateTime_DATE_GET_SECOND(object o)
    int PyDateTime_DATE_GET_MICROSECOND(object o)

    # Getters for time (C macros).
    int PyDateTime_TIME_GET_HOUR(object o)
    int PyDateTime_TIME_GET_MINUTE(object o)
    int PyDateTime_TIME_GET_SECOND(object o)
    int PyDateTime_TIME_GET_MICROSECOND(object o)

    # Getters for timedelta (C macros).
    int PyDateTime_DELTA_GET_DAYS(object o)
    int PyDateTime_DELTA_GET_SECONDS(object o)
    int PyDateTime_DELTA_GET_MICROSECONDS(object o)

    # PyDateTime CAPI object.
    PyDateTime_CAPI *PyDateTimeAPI

    void PyDateTime_IMPORT()

# Datetime C API initialization function.
# You have to call it before any usage of DateTime CAPI functions.
cdef inline void import_datetime():
    PyDateTime_IMPORT

# Create date object using DateTime CAPI factory function.
# Note, there are no range checks for any of the arguments.
cdef inline object date_new(int year, int month, int day):
    return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)

# Create time object using DateTime CAPI factory function
# Note, there are no range checks for any of the arguments.
cdef inline object time_new(int hour, int minute, int second, int microsecond, object tz):
    return PyDateTimeAPI.Time_FromTime(hour, minute, second, microsecond, tz, PyDateTimeAPI.TimeType)

# Create datetime object using DateTime CAPI factory function.
# Note, there are no range checks for any of the arguments.
cdef inline object datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz):
    return PyDateTimeAPI.DateTime_FromDateAndTime(year, month, day, hour, minute, second, microsecond, tz, PyDateTimeAPI.DateTimeType)

# Create timedelta object using DateTime CAPI factory function.
# Note, there are no range checks for any of the arguments.
cdef inline object timedelta_new(int days, int seconds, int useconds):
    return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)

# More recognizable getters for date/time/datetime/timedelta.
# There are no setters because datetime.h hasn't them.
# This is because of immutable nature of these objects by design.
# If you would change time/date/datetime/timedelta object you need to recreate.

# Get tzinfo of time
cdef inline object time_tzinfo(object o):
    if (<PyDateTime_Time*>o).hastzinfo:
        return <object>(<PyDateTime_Time*>o).tzinfo
    else:
        return None

# Get tzinfo of datetime
cdef inline object datetime_tzinfo(object o):
    if (<PyDateTime_DateTime*>o).hastzinfo:
        return <object>(<PyDateTime_DateTime*>o).tzinfo
    else:
        return None

# Get year of date
cdef inline int date_year(object o):
    return PyDateTime_GET_YEAR(o)

# Get month of date
cdef inline int date_month(object o):
    return PyDateTime_GET_MONTH(o)

# Get day of date
cdef inline int date_day(object o):
    return PyDateTime_GET_DAY(o)

# Get year of datetime
cdef inline int datetime_year(object o):
    return PyDateTime_GET_YEAR(o)

# Get month of datetime
cdef inline int datetime_month(object o):
    return PyDateTime_GET_MONTH(o)

# Get day of datetime
cdef inline int datetime_day(object o):
    return PyDateTime_GET_DAY(o)

# Get hour of time
cdef inline int time_hour(object o):
    return PyDateTime_TIME_GET_HOUR(o)

# Get minute of time
cdef inline int time_minute(object o):
    return PyDateTime_TIME_GET_MINUTE(o)

# Get second of time
cdef inline int time_second(object o):
    return PyDateTime_TIME_GET_SECOND(o)

# Get microsecond of time
cdef inline int time_microsecond(object o):
    return PyDateTime_TIME_GET_MICROSECOND(o)

# Get hour of datetime
cdef inline int datetime_hour(object o):
    return PyDateTime_DATE_GET_HOUR(o)

# Get minute of datetime
cdef inline int datetime_minute(object o):
    return PyDateTime_DATE_GET_MINUTE(o)

# Get second of datetime
cdef inline int datetime_second(object o):
    return PyDateTime_DATE_GET_SECOND(o)

# Get microsecond of datetime
cdef inline int datetime_microsecond(object o):
    return PyDateTime_DATE_GET_MICROSECOND(o)

# Get days of timedelta
cdef inline int timedelta_days(object o):
    return (<PyDateTime_Delta*>o).days

# Get seconds of timedelta
cdef inline int timedelta_seconds(object o):
    return (<PyDateTime_Delta*>o).seconds

# Get microseconds of timedelta
cdef inline int timedelta_microseconds(object o):
    return (<PyDateTime_Delta*>o).microseconds



# http://pubs.opengroup.org/onlinepubs/009695399/basedefs/sys/time.h.html

from posix.types cimport suseconds_t, time_t, clockid_t, timer_t
from posix.signal cimport sigevent

cdef extern from "<sys/time.h>" nogil:
    enum: CLOCK_REALTIME
    enum: TIMER_ABSTIME
    enum: CLOCK_MONOTONIC

    # FreeBSD-specific clocks
    enum: CLOCK_UPTIME
    enum: CLOCK_UPTIME_PRECISE
    enum: CLOCK_UPTIME_FAST
    enum: CLOCK_REALTIME_PRECISE
    enum: CLOCK_REALTIME_FAST
    enum: CLOCK_MONOTONIC_PRECISE
    enum: CLOCK_MONOTONIC_FAST
    enum: CLOCK_SECOND

    # Linux-specific clocks
    enum: CLOCK_PROCESS_CPUTIME_ID
    enum: CLOCK_THREAD_CPUTIME_ID
    enum: CLOCK_MONOTONIC_RAW
    enum: CLOCK_REALTIME_COARSE
    enum: CLOCK_MONOTONIC_COARSE
    enum: CLOCK_BOOTTIME
    enum: CLOCK_REALTIME_ALARM
    enum: CLOCK_BOOTTIME_ALARM

    enum: ITIMER_REAL
    enum: ITIMER_VIRTUAL
    enum: ITIMER_PROF

    cdef struct timezone:
        int tz_minuteswest
        int dsttime

    cdef struct timeval:
        time_t      tv_sec
        suseconds_t tv_usec

    cdef struct timespec:
        time_t tv_sec
        long   tv_nsec

    cdef struct itimerval:
        timeval it_interval
        timeval it_value

    cdef struct itimerspec:
        timespec it_interval
        timespec it_value

    int nanosleep(const timespec *, timespec *)

    int getitimer(int, itimerval *)
    int gettimeofday(timeval *tp, timezone *tzp)
    int setitimer(int, const itimerval *, itimerval *)

    int clock_getcpuclockid(pid_t, clockid_t *)
    int clock_getres(clockid_t, timespec *)
    int clock_gettime(clockid_t, timespec *)
    int clock_nanosleep(clockid_t, int, const timespec *, timespec *)
    int clock_settime(clockid_t, const timespec *)

    int timer_create(clockid_t, sigevent *, timer_t *)
    int timer_delete(timer_t)
    int timer_gettime(timer_t, itimerspec *)
    int timer_getoverrun(timer_t)
    int timer_settime(timer_t, int, const itimerspec *, itimerspec *)
